#!/usr/bin/env python3
#-*- coding: utf-8 -*-

import multiprocessing
import os
import socket
import threading
import uuid

import raisin


"""
un lien interressant:
https://www.neuralnine.com/tcp-chat-in-python/
"""

class _Receiver:
    """
    Permet d'etre en ecoute sans bloquer le fil principal.
    Permet donc d'ecouter plusieurs ip a la fois
    """
    def __init__(self, tcp_socket):
        self.tcp_socket = tcp_socket
        self.must_die = False

    def stop(self):
        """
        demande gentiment au server de s'arretter
        """
        self.must_die = True

    def kill(self):
        """
        arrete le serveur imediatement, meme si une opperation est en cours
        """
        if self.is_alive():
            self.join(timeout=0)

    def __iter__(self):
        """
        c'est ici que l'on cede les nouvelles connections
        """
        while not self.must_die:
            yield self.tcp_socket.accept()

class Server:
    """
    c'est un serveur ipv4 et ipv6
    qui permet de gerer plein de clients.
    Il comunique avec un orchestrateur pour gerer les differentes taches demandees par
    les clients
    """
    def __init__(self, signature=None):
        self.signature = signature
        with raisin.Printer("Initialisation of TCP servers...", signature=self.signature) as p:
            
            # partie serveur
            self.port = raisin.worker.configuration.load_settings()["server"]["port"]
            self.listen = raisin.worker.configuration.load_settings()["server"]["listen"]
            myid = raisin.Id()
            self.ipv4 = myid.ipv4_lan
            self.ipv6 = myid.ipv6
            if self.ipv4 == None and self.ipv6 == None:
                raise RuntimeError("Impossible de detecter l'ip, peut etre n'y a-t-il pas internet?")

            if self.ipv4:
                self.tcp_socket_ipv4 = socket.socket(
                    socket.AF_INET,         # socket internet plutot qu'un socket unix
                    socket.SOCK_STREAM)     # creation d'un TCP/IP socket, SOCK_STREAM=>TCP
                self.tcp_socket_ipv4.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1) # on tente de reutiliser le port si possible
                self.tcp_socket_ipv4.bind((str(self.ipv4), self.port))
                self.tcp_socket_ipv4.listen()
                self.receiver_ipv4 = _Receiver(self.tcp_socket_ipv4)
            else:
                self.tcp_socket_ipv4 = None
                self.receiver_ipv4 = None
            if self.ipv6:
                self.tcp_socket_ipv6 = socket.socket(
                    socket.AF_INET6,        # socket internet en ipv6
                    socket.SOCK_STREAM)     # creation d'un TCP/IP socket, SOCK_STREAM=>TCP
                self.tcp_socket_ipv6.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
                self.tcp_socket_ipv6.bind((str(self.ipv6), self.port))
                self.tcp_socket_ipv6.listen()
                self.receiver_ipv6 = _Receiver(self.tcp_socket_ipv6)
            else:
                self.tcp_socket_ipv6 = None
                self.rceiver_ipv6 = None
            
            # partie environement
            self.clients = {}               # c'est tous les clients connectes

    def is_accepted(self, public_key):
        """
        renvoi True si ce client a le droit de se connecter
        retourne False si il n'a pas le droit.
        retourne None si on doit demander l'avis de l'utilisateur 
        Ne pose aucune question a l'utilisateur, retourne sans attendre
        """
        with raisin.Printer("Testing if a pecific client is accepted...", signature=self.signature) as p:
            blacklisted = os.path.join(os.path.expanduser("~"), ".raisin/blacklisted.py")
            if os.path.exists(blacklisted):
                with open(blacklisted, "r", encoding="utf-8") as f:
                    if public_key.replace(b"\n", b"") in eval(f.read()):
                        p.show("It is accepted.")
                        return False
            whitelisted = os.path.join(os.path.expanduser("~"), ".raisin/whitelisted.py")
            if os.path.exists(whitelisted):
                with open(whitelisted, "r", encoding="utf-8") as f:
                    if public_key.replace(b"\n", b"") in eval(f.read()):
                        p.show("It is refoule.")
                        return True
            if raisin.worker.configuration.load_settings()["server"]["accept_new_client"]:
                p.show("It is accepted.")
                return True
            p.show("Maybe, we must ask the question.")
            return None

    def accept(self, identity, signature):
        """
        ajoute le client passe en parametre dans la liste blanche
        ou a la liste noir si l'utilisateur le veut
        Si il faut attendre l'approbation de l'utilisateur,
        cette fonction est bloquante.
        retourne True si au final le client est accepte, et False sinon
        """
        with raisin.Printer("Ask to user for accept a new client...", signature=signature) as p:
            is_a = self.is_accepted(identity["public_key"])
            if is_a == True:
                p.show("The client is already accepted.")
                return True
            if is_a == False:
                p.show("This client is already baned.")
                return False
            question =  "'%s' client shouaite se connecter.\n" % identity["username"] \
                        + "Il travail sur l'os '%s' sur le PC %s.\n" % (identity["os_version"], identity["hostname"]) \
                        + "Il est localise en %s a %s.\n" % (identity["country"], identity["city"]) \
                        + "Il pretend avoir l'adresse mac %s.\n" % identity["mac"]
            rep = raisin.worker.configuration.question_binaire(question, default=None, existing_window=None)
            listed = "whitelisted.py" if rep else "blacklisted.py"
            path = os.path.join(os.path.expanduser("~"), ".raisin", listed)
            data = set()
            with raisin.Lock(id="listed", timeout=30, signature=signature, locality_degrees=2):
                if os.path.exists(path):
                    with open(path, "r", encoding="utf-8") as f:
                        data = eval(f.read())
                data.add(identity["public_key"].replace(b"\n", b""))
                with open(path, "w", encoding="utf-8") as f:
                    f.write(repr(data))
            if rep:
                p.show("The client is desormais accepted.")
            else:
                p.show("This client is desormais baned.")
            return rep

    def close(self):
        """
        permet de fermer proprement les sockets
        """
        with raisin.Printer("Close connection...", signature=self.signature):
            if self.tcp_socket_ipv4:
                self.receiver_ipv4.kill()
                self.tcp_socket_ipv4.close()
            if self.tcp_socket_ipv6:
                self.receiver_ipv4.kill()
                self.tcp_socket_ipv6.close()
            for client in self.clients:
                client["socket"].close()

    def receive(self):
        """
        generateur qui cede les variables:
        client_socket, (ip_client, port_client)
        des l'ors qu'un client tente de se connecter,
        que se soit en ipv4 ou en ipv6
        """
        if self.ipv4 and self.ipv6:
            message = "aux adresses '%s' et '%s'" % (self.ipv4, self.ipv6)
        else:
            message = "a l'adresse '%s'" % (self.ipv4 if self.ipv4 else self.ipv6)
        with raisin.Printer("En ecoute sur le port %d %s..." % (self.port, message), signature=self.signature) as p:
            '''
            en ipv6:
            (   <
                    socket.socket fd=6,
                    family=AddressFamily.AF_INET6,
                    type=SocketKind.SOCK_STREAM,
                    proto=0,
                    laddr=('2a01:e0a:392:b810:ec:6034:b3de:5afa', 1025, 0, 0),
                    raddr=('2a01:e0a:392:b810:ec:6034:b3de:5afa', 52332, 0, 0)
                >,
                ('2a01:e0a:392:b810:ec:6034:b3de:5afa', 52332, 0, 0)
            )
            en ipv4:
            (   <
                    socket.socket fd=5,
                    family=AddressFamily.AF_INET,
                    type=SocketKind.SOCK_STREAM,
                    proto=0,
                    laddr=('192.168.0.48', 1025),
                    raddr=('192.168.0.48', 35758)
                >,
                ('192.168.0.48', 35758)
            )
            '''
            yield from raisin.tools.MergeGenerators(self.receiver_ipv4, self.receiver_ipv6)

    def handle(self, client_socket, key, signature):
        """
        cette methode doit etre lancee dans un thread
        c'est ici que l'on interragit avec un client
        """
        while 1:
            try:
                # Traitement des messages
                with raisin.Printer("Waiting for a reponse...", signature=signature) as p:
                    data = raisin.communication.answering.receive_data(client_socket, signature=signature)
                    if data[0] == 0: # si ce qui suit est un nom de ficher
                        p.show("Reception d'un gros message.")
                        pass
                    elif data[0] == 1: # si ce qui suit et un objet serialise
                        p.show("Reception d'un petit message.")
                        request = raisin.deserialize(
                            data[1:],
                            parallelization_rate=0,
                            psw=None,
                            signature=signature)
                        pass
                    else: # ce cas ne devrai jamais arriver
                        p.show("Le message recut ne respecte pas la norme.")
                        raise ValueError("Le message recut ne respecte pas la norme.")
            except Exception as e:
                # Suppresion est extermination du client
                client_socket.close()
                del self.clients[key]
                break

    def select(self):
        """
        Gere l'acceptation des nouveau clients.
        Si le client est accepte, il est ajouter dans les clients connectes
        et un nouveau thread apparait rien que pour lui
        """
        for client_socket, inf in self.receive(): # on ne fait pas directement
            ip_client = inf[0]                     # client_socket, (ip_client, port_client)
            port_client = inf[1]                    # car le deuxieme tuple contient parfois 4 elements
            with raisin.Printer("Un client d'ip %s se connecte via le port %d..." % (ip_client, port_client)) as p:
                if len(self.clients) >= raisin.worker.configuration.load_settings()["server"]["listen"]: # si il y a deja trop de monde
                    p.show("Client expulse car il y a deja trop de monde connecte.")
                    answer = {"type": "error", "message": "Desole, il y a deja trop de monde."}
                    raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                    continue
                with raisin.Printer("Demande de plus amples informations...", signature=self.signature):
                    question1 = {"type": "question", "question": "How are you?"}
                    raisin.communication.answering.send_object(client_socket, question1, signature=self.signature)
                    try:
                        identity = raisin.communication.answering.receive_object(client_socket, signature=self.signature, timeout=30)
                    except socket.timeout:
                        p.show("Client expulse car il met trop de temps a repondre.")
                        answer = {"type": "error", "message": "Tu es trop lent a repondre."}
                        raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                        continue
                    with raisin.Printer("Verification de la reponse...", signature=self.signature):
                        champs = {
                            "type": str,
                            "public_key": bytes,
                            "username": str,
                            "os_version": str,
                            "hostname": str,
                            "country": str,
                            "city": str,
                            "mac": str,
                            }
                        for clef, genre in champs.items():
                            if clef not in identity:
                                p.show("Client expulse car il n'y a pas de champs '%s' dans sa reponse." % clef)
                                answer = {"type": "error", "message": "Il doit y avoir un champs '%s', ce n'est pas le cas." % clef}
                                raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                                continue
                            if type(identity[clef]) is not genre:
                                p.show("Client expulse car la valeur associee au champs '%s' n'est pas de type '%s'." % (clef, genre))
                                answer = {"type": "error", "message": "N'essai meme pas de hacker le champs '%s'"}
                                raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                                continue
                        if type(identity) is not dict:
                            p.show("Client expulse car il n'a pas repondu avec un dictionaire.")
                            answer = {"type": "error", "message": "Vous devez envoyer un dictionnaire, pas un %s." % type(identity)}
                            raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                            continue
                        if identity["type"] != "identity":
                            p.show("Client expulse car il ne veut pas donner son identite.")
                            answer = {"type": "error", "message": "Le type de reponse atendu est 'identity', pas %s." % str(identity["type"])}
                            raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                            continue
                        if not raisin.re.search(r"-----BEGIN PUBLIC KEY-----(.|\n)+-----END PUBLIC KEY-----", identity["public_key"].decode()):
                            p.show("Client expulse car la clef publique n'est pas au format 'PEM'.")
                            answer = {"type": "error", "message": "La clef publique n'est pas au format 'PEM'."}
                            raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                            continue
                        p.show("Le client a livre son identite.")
                    if raisin.worker.configuration.load_settings()["server"]["force_authentication"]:
                        with raisin.Printer("Envoi d'un challenge pour confirmer l'identite...", signature=self.signature):
                            challenge = uuid.uuid4().bytes   # creation aleatoire d'une phrase 
                            question2 = {                    # on chiffre cette phrase avec la clef publique du client
                                "type":"challenge",
                                "challenge":raisin.worker.security.encrypt_rsa(
                                    challenge,
                                    identity["public_key"],
                                    parallelization_rate=0,
                                    signature=self.signature)}
                            answer_challenge = raisin.communication.answering.send_object(client_socket, question2, signature=self.signature) # on demande au client de la dechiffrer
                            if type(answer_challenge) is not dict:
                                p.show("Client expulse car il n'a pas repondu au challenge avec un dictionaire.")
                                answer = {"type":"error", "message":"Vous devez envoyer un dictionnaire, pas un %s." % type(answer_challenge)}
                                raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                                continue
                            if "type" not in answer_challenge:
                                p.show("Client expulse car il n'y a pas de champ 'type' dans ca reponse.")
                                answer = {"type":"error", "message":"Il doit y avoir un champs 'type' dans votre reponse."}
                                raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                                continue
                            if answer_challenge["type"] != "challenge_back":
                                p.show("Client expulse car ca reponse n'est pas un retour de challenge.")
                                answer = {"type":"error", "message":"Le type de reponse atendu est 'challenge_back', pas %s." % str(answer_challenge["type"])}
                                raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                                continue
                            if "challenge" not in answer_challenge:
                                p.show("Client expulse car il n'a pas voulut repondre au challenge.")
                                answer = {"type":"error", "message":"Il doit y avoir un champs 'challenge' dans votre reponse."}
                                raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                            if type(answer_challenge["challenge"]) is not bytes: # cela evite de pouvoir hacker la methode compare
                                p.show("Client expulse car il a tente de hacker la reponse du chalenge.")
                                answer = {"type":"error", "message":"Cherche pas a me hacker!"}
                                raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                            if answer_challenge["challenge"] != challenge:
                                p.show("Client expulse car il n'a pas reussi le defit.")
                                answer = {"type":"error", "message":"Vous etes un escrot, vous n'avez pas reussi le defit!"}
                                raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                                continue
                            p.show("Le client a reussi le defit.")
                    autorization = self.is_accepted(identity["public_key"])
                    if autorization == None: # si il faut demander l'avis
                        threading.Thread(target=self.accept, args=(identity.copy(), uuid.uuid4().hex)).start() # on pose la question a l'utilisateur
                        p.show("Client expulse car l'utilisateur n'a pas encore donne son accord.") # mais on attend pas qu'il reponde
                        answer = {"type": "error", "message": "Il faut attendre l'avis de l'utilisateur."}
                        raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                        continue
                    elif autorization == False:  # si il faut le virer
                        p.show("Client expulse car il n'a pas le droit de metre les pieds ici!")
                        answer = {"type": "error", "message": "Tu n'a pas le droit de metre les pieds ici!"}
                        raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                        continue
                    if identity["public_key"].replace(b"\n", b"") in self.clients:
                        p.show("Client expulse car il est deja connecte.")
                        answer = {"type": "error", "message": "Tu es deja connecte."}
                        raisin.communication.answering.send_object(client_socket, answer, signature=self.signature)
                        continue
                    p.show("Client accepte.")
                with raisin.Printer("Ajout de ce client dans le serveur...", signature=self.signature):
                    key = identity["public_key"].replace(b"\n", b"")
                    self.clients[key] = {
                        "username": identity["username"],
                        "os_version": identity["os_version"],
                        "hostname": identity["hostname"],
                        "country": identity["counrty"],
                        "city": identity["city"],
                        "mac": identity["mac"],
                        "ip": ip_client,
                        "port": port,
                        "socket": client_socket,
                        "thread": threading.thread(
                            target=self.handle,
                            args=(client_socket, key, uuid.uuid4().hex)),
                    }
                    self.clients[key]["thread"].start()

    def start(self):
        """
        c'est la premiere methode qui doit
        etre appelle apres la creation de l'objet
        """
        return self.select()

    def __del__(self):
        self.close()




#!/usr/bin/env python3
#-*- coding: utf-8 -*-

import raisin

def solve_commente(f, *symbols, **flags):
    """
    Algebraically solves equations and systems of equations.

    Currently supported are:
        - polynomial,
        - transcendental
        - piecewise combinations of the above
        - systems of linear and polynomial equations
        - systems containing relational expressions.

    Input is formed as:
    
    * f
        - a single Expr or Poly that must be zero,
        - an Equality
        - a Relational expression
        - a Boolean
        - iterable of one or more of the above

    * symbols (object(s) to solve for) specified as
        - none given (other non-numeric objects will be used)
        - single symbol
        - denested list of symbols
          e.g. solve(f, x, y)
        - ordered iterable of symbols
          e.g. solve(f, [x, y])

    * flags
        'dict'=True (default is False)
            return list (perhaps empty) of solution mappings
        'set'=True (default is False)
            return list of symbols and set of tuple(s) of solution(s)
        'exclude=[] (default)'
            don't try to solve for any of the free symbols in exclude;
            if expressions are given, the free symbols in them will
            be extracted automatically.
        'check=True (default)'
            If False, don't do any testing of solutions. This can be
            useful if one wants to include solutions that make any
            denominator zero.
        'numerical=True (default)'
            do a fast numerical check if ``f`` has only one symbol.
        'minimal=True (default is False)'
            a very fast, minimal testing.
        'warn=True (default is False)'
            show a warning if checksol() could not conclude.
        'simplify=True (default)'
            simplify all but polynomials of order 3 or greater before
            returning them and (if check is not False) use the
            general simplify function on the solutions and the
            expression obtained when they are substituted into the
            function which should be zero
        'force=True (default is False)'
            make positive all symbols without assumptions regarding sign.
        'rational=True (default)'
            recast Floats as Rational; if this option is not used, the
            system containing floats may fail to solve because of issues
            with polys. If rational=None, Floats will be recast as
            rationals but the answer will be recast as Floats. If the
            flag is False then nothing will be done to the Floats.
        'manual=True (default is False)'
            do not use the polys/matrix method to solve a system of
            equations, solve them one at a time as you might "manually"
        'implicit=True (default is False)'
            allows solve to return a solution for a pattern in terms of
            other functions that contain that pattern; this is only
            needed if the pattern is inside of some invertible function
            like cos, exp, ....
        'particular=True (default is False)'
            instructs solve to try to find a particular solution to a linear
            system with as many zeros as possible; this is very expensive
        'quick=True (default is False)'
            when using particular=True, use a fast heuristic instead to find a
            solution with many zeros (instead of using the very slow method
            guaranteed to find the largest number of zeros possible)
        'cubics=True (default)'
            return explicit solutions when cubic expressions are encountered
        'quartics=True (default)'
            return explicit solutions when quartic expressions are encountered
        'quintics=True (default)'
            return explicit solutions (if possible) when quintic expressions
            are encountered
        """
    return solver.solve(f, *symbols, **flags)


def solve(f, *symbols, **flags):
    """
    Algebraically solves equations and systems of equations.

    Input is formed as:
    
    * f
        - a single Expr or Poly that must be zero,
        - an Equality
        - a Relational expression
        - a Boolean
        - iterable of one or more of the above

    * symbols (object(s) to solve for) specified as
        - none given (other non-numeric objects will be used)
        - single symbol
        - denested list of symbols
          e.g. solve(f, x, y)
        - ordered iterable of symbols
          e.g. solve(f, [x, y])
    * flags
        'dict'=True (default is False)
            return list (perhaps empty) of solution mappings
        'set'=True (default is False)
            return into maybe a list of dictionary, the solution like a raisin.sympy interval
        'exclude=[] (default)'
            don't try to solve for any of the free symbols in exclude;
            if expressions are given, the free symbols in them will
            be extracted automatically.
        'check=True (default)'
            If False, don't do any testing of solutions. This can be
            useful if one wants to include solutions that make any
            denominator zero.
        'numerical=True (default)'
            do a fast numerical check if ``f`` has only one symbol.
        'minimal=True (default is False)'
            a very fast, minimal testing.
        'warn=True (default is False)'
            show a warning if checksol() could not conclude.
        'simplify=True (default)'
            simplify all but polynomials of order 3 or greater before
            returning them and (if check is not False) use the
            general simplify function on the solutions and the
            expression obtained when they are substituted into the
            function which should be zero
        'force=True (default is False)'
            make positive all symbols without assumptions regarding sign.
        'rational=True (default)'
            recast Floats as Rational; if this option is not used, the
            system containing floats may fail to solve because of issues
            with polys. If rational=None, Floats will be recast as
            rationals but the answer will be recast as Floats. If the
            flag is False then nothing will be done to the Floats.
        'manual=True (default is False)'
            do not use the polys/matrix method to solve a system of
            equations, solve them one at a time as you might "manually"
        'implicit=True (default is False)'
            allows solve to return a solution for a pattern in terms of
            other functions that contain that pattern; this is only
            needed if the pattern is inside of some invertible function
            like cos, exp, ....
        'particular=True (default is False)'
            instructs solve to try to find a particular solution to a linear
            system with as many zeros as possible; this is very expensive
        'quick=True (default is False)'
            when using particular=True, use a fast heuristic instead to find a
            solution with many zeros (instead of using the very slow method
            guaranteed to find the largest number of zeros possible)
        'cubics=True (default)'
            return explicit solutions when cubic expressions are encountered
        'quartics=True (default)'
            return explicit solutions when quartic expressions are encountered
        'quintics=True (default)'
            return explicit solutions (if possible) when quintic expressions
            are encountered
    """
    if flags.get("set", False):
        return raisin.geometry.to_set(raisin.sympy.solve(f, *symbols, **flags))
    return raisin.sympy.solve(f, *symbols, **flags)






#!/usr/bin/env python3
#-*- coding: utf-8 -*-


"""
Ce module est charge de repondre aux clients.
Il permet aussi la gestion bas niveau des sockets TCP
afin de serialiser/deserialiser les requettes et les reponsses.
"""

import raisin

def send_data(s, generator, signature):
    """
    permet d'envoyer des donnees
    'generator' est la sortie d'une fonctions de serialisation
    """
    def standardize_generator_sizes(generator, buff_size):
        """
        'generator' est un generateur qui cede des paquets de taille tres variable
        les paquets doivent etre de type 'bytes'
        le but est ici, d'uniformiser la taille des packets afin de renvoyer des packets de 
        'buff_size' octets
        cede donc les paquets au fure et a meusure
        """
        pack = b""
        for data in generator:                                          # on va lentement vider le generateur
            pack += data                                                # pour stocker peu a peu les paquets dans cette variable
            while len(pack) >= buff_size:                               # si le packet est suffisement gros
                yield pack[:buff_size]                                  # on le retourne avec la taille reglementaire
                pack = pack[buff_size:]                                 # puis on le racourci et on recomence le test
        if pack:
            yield pack       
    
    def anticipator(generator):
        """
        cede les packets du generateur accompagne d'un boolean
        qui vaut True si l'element itere est le dernier
        et false sinon. Le generateur doit donc etre capable de ceder au moin un paquet
        """
        actuel = next(generator)
        for pack in generator:
            yield False, actuel
            actuel = pack
        yield True, actuel
    
    with raisin.Printer("Envoi des donnees...", signature=signature) as p:
        for is_end, data in anticipator(standardize_generator_sizes(generator, 1024*1024 -1)):
            p.show("Contenu: %s" % (bytes([is_end]) + data))
            s.send(bytes([is_end]) + data)

def send_object(s, obj, signature):
    """
    fonctionne comme send_data
    mais prend un objet en entree
    n'est efficace que pour les petits objets
    """
    with raisin.Printer("Envoi d'un objet...", signature=signature):
        return send_data(s,
            raisin.serialize(
                obj,
                compresslevel=0,
                parallelization_rate=0,
                copy_file=False,
                signature=signature),
            signature=signature)

def receive_data(s, signature, timeout=None):
    """
    permet de receptionner les donnees,
    meme si elle sont nombreuses
    retourne soit directement les donnees soit le nom du fichier qui les contients
    ce qui permet de distinguer les 2, c'est le premier octet qui vaut 1 si les donnees sont directement recuperee
    ou 0 si ce qui suit est le nom de fichier
    """
    with raisin.Printer("Reception des donnees brutes...", signature=signature) as p:
        default_timeout = s.gettimeout() # peu renvoyer None si il n'y a pas de timeout
        s.settimeout(timeout)
        data = s.recv(1024*1024)
        is_end = data[0]
        if is_end:
            p.show("Contenu: %s" % (b"\x01" + data[1:]))
            s.settimeout(default_timeout)
            return b"\x01" + data[1:]
        filename = os.path.join(str(raisin.temprep), str(uuid.uuid4()))
        with open(filename, "wb") as f:
            f.write(data[1:])
            while not is_end:
                data = s.recv(1024*1024)
                f.write(data[1:])
                is_end = data[0]
        p.show("Contenu: %s" % (b"\x00" + filename.encode("utf-8")))
        s.settimeout(default_timeout)
        return b"\x00" + filename.encode("utf-8")

def receive_object(s, signature, timeout=None):
    """
    fonction comme receive_data mais retourne directement l'objet
    deserialise
    """
    with raisin.Printer("Reception et mise en forme des donnees...", signature=signature):
        data = receive_data(s, signature, timeout=None)
        if data[0] == 1: # dans le cas ou les donnees sont toute presentes
            return raisin.deserialize(data[1:], parallelization_rate=0, signature=signature)
        elif data[0] == 0: # si c'est ecrit dans un fichier
            with open(data[1:].decode("utf-8")) as f:
                obj = raisin.deserialize(f, parallelization_rate=0, signature=signature)
            os.remove(data[1:].decode("utf-8"))
            return obj



def answering(data, signature):
	"""
	permet de traiter une question
	Pour une question de paralelisation, la requette 'data'
	doit etre etre serialise et commencer par b'\x01'
	ou bien doit etre serialise dans un fichier accessible
	retourne le resultat serialise sous la forme d'un generateur
	affin d'offir la possibilite d'une compression dynamique
	"""
	assert type(data) is bytes, "'data' doit etre de type bytes, pas %s." % type(data)
	
	# deserialisation de la requete
	is_direct = data[0]
	if is_direct:
		request = raisin.deserialize(data[1:], parallelization_rate=0, psw=None, signature=signature)
	else:
		with open(data[1:].decode("utf-8"), "rb") as f:
			request = raisin.deserialize(f, parallelization_rate=0, psw=None, signature=signature)
		os.remove(data[1:].decode("utf-8"))

	# traitement de la requette
	reply = "Tiens! voila ta reponse!"

	# emission de la reponsse
	return raisin.serialize(reply, signature=None, buff=1048576, compresslevel=0, copy_file=True, parallelization_rate=0, psw=None)





#!/usr/bin/env python3

"""
Chiffre les donnees avec AES et RSA.
"""

from Cryptodome.Cipher import AES
from Cryptodome.Cipher import PKCS1_OAEP
from Cryptodome.PublicKey import RSA
import hashlib
import itertools
import os
import re
import secrets

import raisin
from .constants import *
from .tools import *
import raisin.serialization.serialize as serialize
from ..tools import Printer


def cipher(data, *, psw=None, compact=1):
    r"""
    |===============================================|
    | Chiffre Les donnees avec l'algotithme adapte. |
    |===============================================|

    Compatible avec decipher et decipher_generator.
    Ne fait aucune verification.
    
    parametres
    ----------
    :param data: Les donnees a chiffrer.
    :type data: bytes
    :param psw: Le mot de passe ou la clef publique rsa.
    :type psw: str, bytes, None
        str: Chiffrage rapide avec AES.
        bytes: Clef publique RSA, Chiffrement lent.
        None: Chiffrage qui utilise la clef publique de raisin.
    :param compact: Determine si les objets serializes.
        doivent restes humainement visibles.
    :type compact: int
        0: Le plus lisible possible.
        1: Le plus compact possible.

    sortie
    ------
    :return: Les donnees chiffrees.
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> from raisin.serialization.cipher import cipher
    >>>
    >>> cipher(b"message en clair", psw="mot de passe")
    b'\x02\xb0\xe9\xe3\xd4\xcdv\xaf\xbf\xde"\xda\xf6@
    \x04\x0fr\x1f\n-Om\x9a\x1e\x03AN\x15\r\x8ei\xbc\xac
    \xe9\xd1\xb7\x0e\xe6-N4n\xb4\x89\x12h.\xfc\xee\xb6'
    >>>
    """
    if psw == None:
        import raisin.security
        psw = raisin.security.get_public_key()

    with Printer("Ciphering datas..."):
        return b"".join(cipher_gen(
            to_gen(pack=data),
            psw=psw,
            compact=compact
            ))

def cipher_aes(data, passphrase, compact):
    r"""
    |============================================|
    | Chiffre des donnees avec l'algorithme AES. |
    |============================================|

    Ne fait aucune verifications sur les entrees.

    parametres
    ----------
    :param data: Donnees brute non chifrees.
    :type data: bytes
    :param passphrase: Sorte de mot de passe sur 32 octets.
    :type passphrase: bytes
    :param compact: Determine si les objets serializes.
        doivent restes humainement visibles.
    :type compact: int
        0: Le plus lisible possible.
        1: Le plus compact possible.

    sortie
    ------
    :return: les donnees chifrees.
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> import hashlib
    >>> from raisin.serialization.cipher import cipher_aes
    >>>
    >>> passphrase = hashlib.blake2s(b'mot de passe', digest_size=32).digest()
    >>> cipher_aes(b"message en clair", passphrase)
    b"oz4\x0bE\xeaD\xb5\x8d\x87\xads\xcf\x96\xaa\xeb
    \xce ?\x10)\xd2\xf6'\xe1\xcfk*\x06\x1a\xe8b\xe2<
    \x15%\x9d\xb9}7\xa8\xe02j\xab\x04\x97\x94"
    >>>
    """
    with Printer("Ciphering bloc with AES..."):
        cipher = AES.new(passphrase, AES.MODE_EAX)
        ciphertext, tag = cipher.encrypt_and_digest(data)
        return HEADER["aes bloc"][compact] + cipher.nonce + tag + ciphertext

def cipher_rsa(data, public_key_pem, compact):
    r"""
    |============================================|
    | Chiffre des donnees avec l'algorithme RSA. |
    |============================================|
    
    Attention, les donnees chiffrees prennent 2 fois plus de place.
    N'effectue pas de verifications sur les entrees.

    parametres
    ----------
    :param data: Donnees brute non chiffrees.
    :type data: bytes
    :param public_key_pem: Clef publique serialisee au format PEM.
        La clef doit etre dechifree.
    :type public_key_pem: bytes
    :param compact: Determine si les objets serializes.
        doivent restes humainement visibles.
    :type compact: int
        0: Le plus lisible possible.
        1: Le plus compact possible.

    sortie
    ------
    :return: les donnees chiffrees 
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> from raisin.security import rsa_keys
    >>> from raisin.serialization.cipher import cipher_rsa
    >>> priv, pub = rsa_keys()
    >>> cipher_rsa(b"message en clair", pub)
    b'\x08\x80B+\xc3"\xe9z\x07 ... \x89\xaf\xfe\xfd0\xf1\x87'
    >>>
    """
    with Printer("Ciphering bloc with RSA..."):
        encryptor = PKCS1_OAEP.new(RSA.importKey(public_key_pem)) # deserialisation de la clef RSA
        return HEADER["rsa bloc"][compact] + b"".join(
            concat_gen([
                encryptor.encrypt(data[i:i+500])
                for i in range(0, len(data), 500)
                ])
            )

def cipher_gen(gen, psw, compact):
    r"""
    |=============================================|
    | Chiffre sequentiellement chaque packets     |
    | cedes par le generateur sans les regrouper. |
    |=============================================|
    
    Ne fait pas de verification sur les entrees.

    parametres
    ----------
    :param generator: Generateur qui cede des 'bytes'.
    :type generator: iter
    :param psw: Le mot de passe ou la clef publique rsa.
    :type psw: str, bytes, None
        str: Chiffrage rapide avec AES.
        bytes: Clef publique RSA, Chiffrement lent.
    :param compact: Determine si les objets serializes.
        doivent restes humainement visibles.
    :type compact: int
        0: Le plus lisible possible.
        1: Le plus compact possible.

    sortie
    ------
    :return: Les donnees Chiffrees.
    :rtype: iter

    exemple
    -------
    :Example:
    >>> import uuid
    >>> from raisin.serialization.cipher import cipher_generator
    >>>
    >>> gen = (uuid.uuid4().bytes for i in range(3))
    >>> for d in cipher_generator(gen):
    ...     print(d)
    ...
    b'\x03\x08\x82\x08\x80\x8d\xb ...
    b"\xb0\xb3J*~\xfc-+\x1 ...
    b"\xb0c\x0fsbT,\xd36\xfeJ ...
    >>>
    """
    with Printer("Ciphering generator..."):
        for i, (is_end, data) in enumerate(anticipate(to_gen(gen=gen))):
            if i == 0 and is_end and len(data) <= 500: # Dans le cas ou l'on a pas grand chose a chiffrer.
                if isinstance(psw, str):
                    sel = secrets.token_bytes(32) # Rend difficile l'attaque d'arc en ciel.
                    passphrase = hashlib.blake2s(
                        sel + psw.encode("utf-8"),
                        digest_size=32
                        ).digest()
                    yield HEADER["aes sel"][compact] \
                        + sel \
                        + cipher_aes(data, passphrase, compact)
                else:
                    yield cipher_rsa(data, psw, compact)
                break
            elif i == 0: 
                if isinstance(psw, str): # Beaucoup a faire en AES.
                    sel = secrets.token_bytes(32) # Sel contre l'arc en ciel.
                    hash_ = hashlib.sha512(sel + psw.encode("utf-8")).digest()
                    passphrases = [hash_[:32], hash_[32:]]
                    yield HEADER["aes gen"][compact] + sel
                else: # Si on doit faire beaucoup en RSA.
                    passphrases = [secrets.token_bytes(32) for _ in range(500//32)]
                    yield HEADER["rsa gen"][compact] + b"".join(
                        concat_gen(
                            frozgen=[cipher_rsa(b"".join(passphrases), psw, compact)]
                            )
                        )
            ciphdata = cipher_aes(data, passphrases[i%len(passphrases)], compact)
            yield size_to_tag(len(ciphdata)) + ciphdata

def _cipher_file(filename, **kwargs):
    """
    |=====================|
    | Chiffre un fichier. |
    |=====================|
    Supprime l'ancien fichier pour ne laisser que le nouveau.
    Ne fait aucune veriication sur les donnees d'entrees.

    parametres
    ----------
    :param filename: Le nom d'un fichier ou son chemin.
    :type filename: str
    
    Les autres arguments sont les memes que
    ceux de la fonction 'cipher_gen'.

    sortie
    ------
    :return: Le nom du fichier chiffre.
    :rtype: str

    exemple
    -------
    :Example:
    >>> from raisin.security.encrypt import _cipher_file
    >>> _cipher_file("monfichier.txt")
    'monfichier.txt.crais'
    >>>
    """
    def generator(filename):
        """
        Cede filename par petits bouts.
        """
        with open(filename, "rb") as f:
            while 1:
                pack = f.read(BUFFER_SIZE)
                if pack == b"":
                    break
                yield pack
    
    with Printer("Encryption of the file %s..." % repr(path)):
        new_file = filename + ".crais"
        try:
            with open(new_file, "wb") as f:
                for pack in cipher_generator(
                        generator(filename),
                        **kwargs):
                    f.write(pack)
            os.remove(filename)
        except Exception as e:
            if os.path.exists(new_file):
                os.remove(new_file)
            raise e from e
        return new_file

def authenticity(gen, compact):
    """
    |==============================================|
    | Calcul est ajoute un hash au contenu de gen. |
    |==============================================|

    Ne fait aucune verifications sur les entrees.

    parametres
    ----------
    :param generator: Generateur qui cede des 'bytes'.
    :type generator: iter
    :param compact: Determine si les objets serializes.
        doivent restes humainement visibles.
    :type compact: int
        0: Le plus lisible possible.
        1: Le plus compact possible.

    sortie
    ------
    :return: Les donnes d'entree suivi du hash
        permetant d'assurer l'authencicite du generateur.
    :rtype: <type generator>
    """
    hash_ = b""
    with Printer("Put authenticity..."):
        yield HEADER["authenticity"][compact]
        for pack in to_gen(gen=gen):
            hash_ = hashlib.sha512(hash_ + pack).digest()
            yield size_to_tag(len(pack)) + pack
        yield size_to_tag(len(hash_)) + hash_




#!/usr/bin/env python3

import copy
import glob
import io
import ipaddress
import json
import os
import platform
import pickle
import re
import shutil
import socket
import sys
import tempfile
import threading
import time
import urllib.request
import uuid

class MergeGenerators:
    """
    permet d'iterer sur plusieur generateurs a la fois.
    Chaque generateur est independament excecuter sur un
    faux thread separe
    """
    class _AssyncGen(threading.Thread):
        def __init__(self, papa, gen):
            threading.Thread.__init__(self)
            self.papa = papa
            self.gen = gen
            self.queue = []
            self.lock = threading.Lock()
            self.start()

        def run(self):
            for element in self.gen:
                with self.lock:
                    self.queue.append(element)
                    try:
                        self.papa.lock.release()
                    except RuntimeError:
                        continue

    def __init__(self, *generators):
        """
        'generators' est un tuple contenant des iterables
        """
        self.lock = threading.Lock()
        self.generators = [MergeGenerators._AssyncGen(self, iter(g)) for g in generators]

    def __iter__(self):
        """
        cede jusqu'au bout le contenu desordonne
        de chaque sous generateur
        """
        while self.generators:  # tant que tout n'est pas termine
            self.lock.acquire() # on attend qu'il se passe quelque-chose
            for i, gen in enumerate(self.generators):
                if gen.queue:
                    with gen.lock:
                        yield from gen.queue
                        gen.queue = []
                if not gen.is_alive():
                    del self.generators[i]

class Id:
    """
    Recupere l'identifiant de cet ordinateur.
    """
    class Version:
        """
        represente une version et permet des comparaisons
        entre les versions
        """
        def __init__(self, version_str):
            self.version_str = version_str
            self.moi_list = self._simplify(list(map(int, self.version_str.split("."))))

        def _simplify(self, version_list):
            """
            retourne une liste simplifier (sans les 0 a la fin)
            """
            assert type(version_list) is list, "'version_list' doit etre de type list, pas %s." % type(version_list)
            while 1:
                if len(version_list):
                    if version_list[-1]:
                        break
                    else:
                        del version_list[-1]
                else:
                    break
            return version_list

        def __repr__(self):
            return self.version_str

        def __eq__(self, other_str):
            """
            renvoi True si 'other_str' correspond a la meme version que self
            """
            assert type(other_str) is str, "La variable 'other' doit etre de type str, non pas %s." % type(other_str)
            other_list = self._simplify(list(map(int, other_str.split("."))))
            return self.moi_list == other_list

        def __gt__(self, other_str):
            """
            retourne True si 'other_str' est plus petit stricetement que self
            """
            assert type(other_str) is str, "La variable 'other' doit etre de type str, non pas %s." % type(other_str)
            other_list = self._simplify(list(map(int, other_str.split("."))))
            for lui, moi in zip(other_list, self.moi_list):
                if lui < moi:
                    return True
                if lui > moi:
                    return False
            return len(other_list) < len(self.moi_list)

        def __lt__(self, other_str):
            return not self > other_str and not self == other_str

        def __ge__(self, other_str):
            return self > other_str or self == other_str

        def __le__(self, other_str):
            return self < other_str or self == other_str

    def __init__(self):
        self.hostname = None
        self.username = None
        self.mac = None         # adresse mac
        self.ipv4_lan = None    # adresse ipv4 du reseau local
        self.ipv4_wan = None    # adresse ipv4 du reseau exterieur
        self.ipv6 = None        # adresses ipv6 publique
        self.country = None     # initiale en minuscule du pays ex: "fr", "en"
        self.city = None        # c'est la ville ex: "grenoble"
        self.bus = None         # 64 ou 32 bits
        self.python_version = None # version de python
        self.os_version = None  # version de l'os qui interprete ce python
        self.system = None      # nom generique du system d'exploitation
        self.has_admin = None   # dit si l'utilisateur a les droits admin
        self.last_connection = 0 # moment de la derniere connection

    def __call__(self):
        """
        Se retourne soit meme apres s'etre mis a jour.
        """
        if self.last_connection == 0:
            self.hostname = self.get_hostname() # on recupere le hostname
            self.username = self.get_username() # on recupere le nom d'utilisateur
            self.bus = 64 if "64" in platform.machine() else 32 if "32" in platform.machine() else None # 64 ou 32 bits
            self.python_version = Id.Version(self.get_python_version()) # version de python
            self.os_version = platform.version() # version de l'os qui interprete ce python
            self.system = platform.system().lower() # nom generique du system d'exploitation
            self.has_admin = self.get_admin() # on regarde les droits que l'on a
        
        if time.time() - self.last_connection > 5*60: # on actualise au maximum toute les 5 minutes
            self.run()              # mise a jour des attribut qui necessitent internet
            self.last_connection = time.time() # moment de la derniere connection

        return self

    def run(self):
        """
        methode lancee automatiquement de facon assynchrone avec
        l'autre methode self.start()
        """
        ti = time.time()
        while time.time()-ti < 120: # on s'autorise 2 minute de latence
            nouveau_mac = self.get_mac() # on recupere l'adresse mac
            self.mac = nouveau_mac if nouveau_mac is not None else self.mac # on met a jour l'adresse mac
            nouveau_ipv4_lan = self.get_ipv4_lan() # on tente de recuperer l'ipv4 locale
            self.ipv4_lan = nouveau_ipv4_lan if nouveau_ipv4_lan is not None else self.ipv4_lan # on et aussi a jour cette adresse ipv4
            nouveau_ipv4_wan = self.get_ipv4_wan() # on tente de recuperer l'ipv4 exterieur
            self.ipv4_wan = nouveau_ipv4_wan if nouveau_ipv4_wan is not None else self.ipv4_wan # on et aussi a jour cette adresse ipv4
            nouveau_ipv6 = self.get_ipv6() # on tente de recuperer l'ipv6 exterieur
            self.ipv6 = nouveau_ipv6 if nouveau_ipv6 is not None else self.ipv6 # on et aussi a jour cette adresse ipv6
            nouveau_country = self.get_country() # on tente de savoir environ ou est l'ordinateur
            self.country = nouveau_country if nouveau_country is not None else self.country # on met a jour seulement si il y a du nouveau
            nouveau_city = self.get_city() # on tente de savoir un peu plus precisement ou se trouve l'ordinateur
            self.city = nouveau_city if nouveau_city is not None else self.city # on met a jour ce parametre
            nouveau_admin = self.get_admin() # on regrade si les droits administrateur on changes
            self.has_admin = nouveau_admin if nouveau_admin else self.has_admin # on met a jour les droits
            if nouveau_ipv4_wan or nouveau_ipv6 or nouveau_country: # si l'on est connecte a internet
                break           # on arrete de chercher
            else:               # si on est pas connecte a internet, on attend 5 secondes
                time.sleep(5)   # avant de retenter

    def update(self):
        """
        met imediatement a jour les toutes les informations
        """
        self.last_connection = 0
        return self()
        
    def get_hostname(self):
        """
        retourne l'identifiant de l'ordinateur
        """
        try:
            return socket.gethostname()
        except socket.error:
            return None

    def get_username(self):
        """
        retourne le nom d'utilisateur
        """
        try:
            return os.environ["USERNAME"]
        except:
            return os.path.basename(os.path.expanduser("~"))

    def get_mac(self):
        """
        retourne l'adress mac du pc
        """
        try:
            return ":".join(["{:02x}".format((uuid.getnode() >> i) & 0xff) for i in range(0, 8*6, 8)][::-1])
        except KeyboardInterrupt as e:
            raise e from e
        except:
            return None

    def get_ipv4_lan(self):
        """
        retourne l'ip sur le reseau local
        """
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)        # un socket en mode UDP (DGRAM)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)     # broadcast
            s.connect(("<broadcast>", 0))                               # le mot-cle python pour INADDR_BROADCAST
            ip_local, port = s.getsockname()                            # renvoi de l'adresse source
            return ipaddress.IPv4Address(ip_local)                      # on retourne donc la vrai adresse
        except socket.error:                                            # mais l'on est pas certain que cette methode fonctionne a tous les coups
            return None                                                 # c'est pour cela que l'on prend des precautions
    
    def get_ipv4_wan(self):
        """
        tente de retourner l'ip d'internet
        """
        @timeout_decorator(30)
        def get_ipify():
            dic = eval(urllib.request.urlopen("https://api.ipify.org/?format=json").read().decode())
            return ipaddress.IPv4Address(dic["ip"])

        @timeout_decorator(30)
        def get_ipinfo():
            dic = eval(urllib.request.urlopen("http://ipinfo.io/json").read().decode())
            return ipaddress.IPv4Address(dic["ip"])

        @timeout_decorator(30)
        def get_icanhazip():
            ip = eval(urllib.request.urlopen("http://icanhazip.com/").read().decode())
            return ipaddress.IPv4Address(ip)

        try:
            return get_ipify()
        except KeyboardInterrupt as e:
            raise e from e
        except:
            try:
                return get_ipinfo()
            except KeyboardInterrupt as e:
                raise e from e
            except:
                try:
                    return get_icanhazip()
                except KeyboardInterrupt as e:
                    raise e from e
                except:
                    return None
    
    def get_country(self):
        """
        retourne les initiales du pays dans lequel on est
        retourne None en cas d'echec de connection a internet
        """
        @timeout_decorator(30)
        def get_ipinfo():
            dic = eval(urllib.request.urlopen("http://ipinfo.io/json").read().decode())
            return dic["country"].lower()

        try:
            return get_ipinfo()
        except KeyboardInterrupt as e:
            raise e from e
        except:
            return None

    def get_city(self):
        """
        retourne les initiales du pays dans lequel on est
        retourne None en cas d'echec de connection a internet
        """
        @timeout_decorator(30)
        def get_ipinfo():
            dic = eval(urllib.request.urlopen("http://ipinfo.io/json").read().decode())
            return dic["city"].lower()

        try:
            return get_ipinfo()
        except KeyboardInterrupt as e:
            raise e from e
        except:
            return None

    def get_ipv6(self):
        """
        tente de retourner l'ipv6 publique
        """
        def get_local():
            result = socket.getaddrinfo(self.hostname, 0, socket.AF_INET6)
            return [ipaddress.IPv6Address(a[4][0]) for a in result]

        @timeout_decorator(30)
        def get_ipify():
            dic = eval(urllib.request.urlopen("https://api6.ipify.org?format=json").read().decode())
            return ipaddress.IPv6Address(dic["ip"])

        @timeout_decorator(30)
        def get_myip():
            ip = urllib.request.urlopen("http://v6.ipv6-test.com/api/myip.php").read().decode()
            return ipaddress.IPv6Address(ip)

        @timeout_decorator(30)
        def get_seeip():
            dic = eval(urllib.request.urlopen("https://ip6.seeip.org/json").read().decode())
            return ipaddress.IPv6Address(dic["ip"])

        try:
            return get_ipify()
        except KeyboardInterrupt as e:
            raise e from e
        except:
            try:
                return get_myip()
            except KeyboardInterrupt as e:
                raise e from e
            except:
                try:
                    return get_seeip()
                except KeyboardInterrupt as e:
                    raise e from e
                except:
                    return None
    
    def get_python_version(self):
        """
        retourne la version de python sur laquelle on est entrain de travailler
        """
        return "%d.%d.%d" % (sys.version_info.major, sys.version_info.minor, sys.version_info.micro)

    def get_admin(self):
        """
        Retourne True si l'utilisateur a les droits administrateur.
        """
        if os.name == "nt":
            try:
                # only windows users with admin privileges can read the C:\windows\temp
                temp = os.listdir(os.sep.join(
                    [os.environ.get("SystemRoot", "C:\\windows"),
                    "temp"]))
            except:
                return False
            else:
                return True
        else:
            if "SUDO_USER" in os.environ and os.geteuid() == 0:
                return True
            return False


    def __repr__(self):
        self.__call__()
        chaine = "{:<14s}".format("username") + ": " + self.username + "\n"    \
                 + "{:<14s}".format("hostname") + ": " + self.hostname + "\n"  \
                 + "{:<14s}".format("has admin") + ": " + ("yes" if self.has_admin else "no") + "\n" \
                 + "{:<14s}".format("city") + ": " + self.city + "\n"          \
                 + "{:<14s}".format("country") + ": " + self.country + "\n"    \
                 + "{:<14s}".format("system") + ": " + self.system + "\n"      \
                 + "{:<14s}".format("python version") + ": " + str(self.python_version) + "\n" \
                 + "{:<14s}".format("bus") + ": " + str(self.bus) + " bits\n"  \
                 + "{:<14s}".format("os") + ": " + self.os_version  + "\n"     \
                 + "{:<14s}".format("mac") + ": " + self.mac + "\n"            \
                 + "{:<14s}".format("ipv4 lan") + ": " + str(self.ipv4_lan) + "\n" \
                 + "{:<14s}".format("ipv4 wan") + ": " + str(self.ipv4_wan) + "\n" \
                 + "{:<14s}".format("ipv6") + ": " + str(self.ipv6)
        return chaine

    def __str__(self):
        return repr(self)

class Lock:
    """
    permet de verouiller bloquer le reste du programe, ou a un autre processus
    """
    def __init__(self, id="default", timeout=60, priority=0, signature=None, display=True, locality_degrees=1):
        """
        'id' (STR) => est l'identifiant de ce verrou, c'est ce qui permet de le rendre independant des autres
        'timeout' (INT, FLOAT) => temps de validite du verrou en s
        'priority' (INT) => l'utilisateur 'n' est prioritaire devant [n-1, n-2, ..., 2, 1, 0] avec n < 1000
        'signature' (obj) => permet si voulu, d'informer l'utilisateur du processus appelant
        'display' (BOOL) => True pour afficher et False pour ne pas appeller afficher
        'locality_degrees (BOLL) => gere la portee de ce verrou:
            -0 => ultra local, visible seulement depuis cette instance du verrou
            -1 => local a ce processus, thread de threading, visible tout de meme depuis differentes instances
            -2 => local a la machine, multiprocessing et plusieur instance du programe
            -3 => visible par tout le LAN, global a tous les ordi present deriere ce meme routeur
            -4 => visible par tous ceux qui appartiennent au reseau de raisin WAN
        """
        self.id = id                                                                        # identifiant de ce verrou
        self.timeout = timeout                                                              # temps de validitee du verrou pose
        self.priority = priority                                                            # est un entier, plus il est elever, plus la prioritee est haute
        self.signature = signature                                                          # permet d'afficher le message au bon endroit
        self.display = display                                                              # True si il faut afficher le message
        self.local = True                                                                   # est True si ce verrou est seulement actif sur ce programe
        self.code = uuid.uuid4().hex                                                        # identifiant unique propre a ce verrou
        if self.local:                                                                      # si le verrou est local, propre a ce processus
            self.path_file = os.path.join(temprep(), "lock_%s.txt" % self.id)               # le chemin qui mene jusqu'au verrou
        else:                                                                               # si seul tout le monde doit pouvoir y avoir acces
            self.path_file = os.path.join(temprep(), "lock_%s.txt" % self.id)               # chemin bien visible par tous

    def acquire(self):
        """
        depose le verrou et retourne une fois qu'il est depose
        """
        while os.path.exists(self.path_file):                                               # tant que le fichier existe
            try:                                                                            # on va essayer de lire ce qu'il y a dedand
                with open(self.path_file, "r") as f:                                        # tentative d'ouverture
                    lock_priority, lock_code, lock_timeout = f.read().split(";")            # et de lecture du fichier
                    lock_priority, lock_code, lock_timeout = int(lock_priority), lock_code, float(lock_timeout) # convertion dans les bon types
                if (lock_code == self.code) and (lock_priority == 999):                     # si on en est plainement maitre
                    return None                                                             # on se sauve vite de la
                elif (lock_code == self.code) and (lock_priority != 999):                   # si personne ne semble se manifester
                    with open(self.path_file, "w") as f:                                    # on va tenter de s'en emparer vraiment
                        f.write("999;"+self.code+";"+str(time.time()+self.timeout))         # en y metant nos empreintes
                elif (lock_code != self.code) and (self.priority > lock_priority):          # si un autre processus depose timidement son verrou mais que l'on est plus fort que lui
                    with open(self.path_file, "w") as f:                                    # on va alors a notre tour tenter de le prendre timidement
                        f.write(str(self.priority)+";"+self.code+";"+str(time.time()+self.timeout)) # en y inscrivant notre volontee
                    if self.priority != 999:
                        time.sleep(0.02)                                                    # on est poli, on laisse la priorite au plus fort
                elif time.time() > lock_timeout:                                            # si le verrou est trop vieu pour avoir encore de la valeur
                    os.remove(self.path_file)                                               # on tente alors de le degomer
                else:                                                                       # dans le cas ou l'on est dans aucune des ces situation, c'est qu'il faut attendre
                    time.sleep(0.01)                                                        # on fait alors une petite pause qui soulage le resources
            except:
                time.sleep(0.3)

        try:                                                                                # si il n'y a pas de fichiers
            with open(self.path_file, "w") as f:                                            # c'est alors possible de prendre la main
                f.write(str(self.priority)+";"+self.code+";"+str(time.time()+self.timeout)) # on essai donc de le faire
            if self.priority != 999:
                pass # pff sinon ca lague trop!
                # time.sleep(0.02)                                                          # on laisse le temps au processus plus fort de se retourner
            return self.acquire()                                                           # on repart donc en haut
        except:                                                                             # si il y a eu une erreur quelqonque
            return self.acquire()                                                           # on retent notre chance une deuxieme fois

    def release(self):
        """
        tente de liberer l'acces
        """
        try:                                                                                #on essai de supprimer le fichier
            os.remove(self.path_file)
        except:
            pass
        return None

    def is_lock(self):
        """
        retourne False si l'acces est libre
        """
        if not os.path.exists(self.path_file):                                              #dans le cas ou il n'y a pas de fichier
            return False                                                                    #c'est que l'acces est libre
        try:                                                                                #dans le cas ou il y a un fichier
            with open(self.path_file, "r") as f:                                            #on va tenter de le lire
                lock_priority, lock_code, lock_timeout = f.read().split(";")                #et de lecture du fichier
                lock_priority, lock_code, lock_timeout = int(lock_priority), lock_code, float(lock_timeout)#convertion dans les bon types
            if time.time() > lock_timeout:                                                  #si le verrou date de gerusalem
                return False                                                                #alor on considere que l'acces est libre
            return True                                                                     #dans le cas contraire, la voix n'est pas libre
        except:                                                                             #si ca capote
            return True                                                                     #en cas d'erreur aussi, la voix n'est pas disponible

    def is_free(self):
        """
        reourne True si l'acces est libre
        """
        return not(self.is_lock())

    def __enter__(self):
        if self.display:
            with Printer("wait "+str(self.id)+" freedom...", signature=self.signature):     #attente de la cess libre
                self.acquire()
        else:
            self.acquire()
        return self

    def __exit__(self, *args):
        if self.display:
            with Printer("release %s..." % self.id, signature=self.signature):               #rendre a nouveau l'acces libre
                self.release()
        else:
            self.release()

class LockDecorator:
    """
    permet de gerer automatiquement un verou l'ors de l'appelle du fonction
    permet aussi de s'assurer que la fonction n'est excecutee qu'une seule fois
    """
    def __init__(self, id, *, timeout=60, priority=0, signature=None, locality_degrees=0, jump=False):
        """
        'jump' (BOOL) => est un boolen qui quand il est a True, fait squiser l'execution de la fonction si elle est verouillee
        """
        self.id = id                                # identifiant du verrou
        self.timeout = timeout                      # temps de validite du verrou en s
        self.priority = priority
        self.signature = signature
        self.locality_degrees = locality_degrees
        assert type(jump) is bool, "'jump' doit etre un booleen, pas un %s." % type(jump)
        self.jump = jump
        
    def __call__(self, func):
        """
        cette methode n'est appelle qu'une seul fois au moment de l'import du programe qui contient
        la fonction decoree
        retourne la fonction decoree
        """
        verrou = Lock(self.id, timeout=self.timeout, priority=self.priority, signature=self.signature, locality_degrees=self.locality_degrees)
        def change_func(*args, **kwargs):
            if self.jump:
                if verrou.is_lock():
                    raise ValueError("Le verrou est deja pris.")
            with verrou:
                return func(*args, **kwargs)

        return change_func

class Printer:
    """
    permet d'afficher joliment les choses dans le terminal
    """
    def __init__(self, text="", *, display=None, signature=None):
        """
        'display' est un entier:
            cet entier correspond au nombre d'indentation maximum
            si il vaut None, la derniere valeur donnee est prise
        'signature' est n'importe quel objet python hashable, il permet de differencier les diferents processus
        pour qu'il n'y ai pas de conflit avec les threads
        """
        assert display is None or type(display) is int, "'display' doit etre un entier ou None, pas %s." % display
        assert type(text) is str, "'text' doit etre une chaine de caractere, pas un %s." % type(text)
        self.text = text
        self.display = display                                                  # s'actualise a l'appel de self.get_indentation
        self.signature = signature if signature != None else os.getpid()

        self.path = os.path.join(temprep(), "printer_infos.pk")
        self.nbr_colones = 1                                                    # s'actualise quand self.get_indentation est appelee, nbr de colone d'affichage dans le terminal
        self.ma_colone = None                                                   # s'actualise au meme moment que 'nbr_colones', rang de la colone en cours

    def get_indentation(self, ecart=0):
        """
        renvoi le nombre d'indentations actuel (sans 'ecart')
        en profite pour enregistrer la nouvelle indentation a venir (avec ecart)
        'ecart' est le nombre d'indentation a ajouter algebriquement dans le fichier,
        0 ne change donc rien
        met aussi en arriere plan, a jour le nbr de colones a venir
        """
        exists = os.path.exists(self.path)                                      # on stocke ce resultat dans une variable afin de gagner du temps
        if ecart == 0 and exists and self.ma_colone:                            # si l'on doit faire de la lecture seule du fichier
            with Lock(id="printer", timeout=5, display=False, locality_degrees=1): # on pose un verrou a part afin de minimiser le temps d'acces au verrou
                with open(self.path, "rb") as f:                                # une fois qu'on a acces au fichier
                    content = pickle.load(f)                                    # on se contente de lire son contenu
            self.display = content[self.signature]["display"]                   # puis on liber tout de suite le verrou
            self.nbr_colones = len(content)
            self.ma_colone = content[self.signature]["colone"]
            return content[self.signature]["indentation"]                       # en faisant la deserialisation et l'analyse hors du verrou
        
        else:                                                                   # si on ne doit pas seulment lire mais qu'on doit aussi mettre a jour l'indentation
            with Lock(id="printer", timeout=10, display=False, locality_degrees=1): # on pose un autre verrou
                if os.path.exists(self.path):                                   # si il y a deja un fichier
                    with open(self.path, "rb") as f:                            # alors on commence par extraire les informations qui s'y trouve
                        content = pickle.load(f)                                # la on fait tout dans le verrou car de toute facon il va faloir reecrir
                else:                                                           # si par contre il n'y a pas deja une configuration
                    content = {}                                                # on edite une dictionaire tout vide
                
                content[self.signature] = content.get(self.signature, {})       # bref, comme on sait pas ce qui y a dans le dico
                content[self.signature]["indentation"] = content[self.signature].get("indentation", 0) + ecart # on prend toutes les precautions qu'on puisse prendre
                self.ma_colone = content[self.signature].get("colone", self.nbr_colones - 1)
                content[self.signature]["colone"] = self.ma_colone
                self.display = content.get(self.signature, {}).get("display", self.display if self.display is not None else 5)
                content[self.signature]["display"] = self.display
                
                indentation = content[self.signature]["indentation"] - ecart
                self.nbr_colones = len(content)
                if not indentation + ecart:                                     # si on l'afficage suivant se fera a ras
                    del content[self.signature]                                 # alors on se supprime afin de laisser de la place aux nouveaux
                    content = {
                        sign:{
                            inf:(val - 1 if inf == "colone" and val > self.ma_colone else val) \
                            for inf,val in infs.items()}
                        for sign, infs in content.items()}                      # puis on met aussi a jour les colones des autres
                with open(self.path, "wb") as f:                                # bref, les donnees sont bien a jour dans la RAM
                    pickle.dump(content, f)                                     # mais il faut encore les ecrire dans le fichier
            return indentation

    def indent(self, text):
        """
        retourne le texte bien indente avec le minimum entre l'indentation courante
        et la valeur maximum d'intentation admissible
        """
        indentation = min(self.get_indentation(), self.display if self.display is not None else 5) # nbr d'indentation a ajouter au texte
        return "\n".join(("\t"*indentation + l for l in text.split("\n")))      # on indentate alors chaque ligne, en prenant en compte les retours a la ligne

    def pprint(self, expression):
        """
        retourne la representation de l'expression avec de beau caractere utf-8
        sur une seule ligne dans la meusure du possible
        en cas d'echec, retourne str(expression)
        """
        if type(expression) is not str and raisin.sympy is not None:
            s = sys.stdout
            with io.StringIO() as f:
                sys.stdout = f
                try:
                    raisin.sympy.pprint(expression)         # on utilise sympy pour ecrire joliment
                    message = f.getvalue()[:-1].split("\n") # ainsi on recupere se qui est ecrit (\n exclu)
                    sys.stdout = s                          # puis  avant qu'il ne se passe trop de chose, on remet tout comme avant
                    if len(message) == 1:                   # si le tout tient sur une ligne
                        return message[0]                   # c'est gagne
                    elif len(message) == 0:                 # si il c'est passe des trucs bresson
                        return str(expression)              # pas de panique, on ignore les magouilles de sympy
                    else:                                   # si le tout ne tien pas sur une seule ligne
                        if type(expression) is dict:        # on essai de moin bourriner
                            return str({self.pprint(c): self.pprint(v) for c, v in expression.items()})# on decompose allors les operations en operation plus simples
                        elif type(expression) is list:
                            return str([self.pprint(v) for v in expression])
                        elif type(expression) is tuple:
                            return str(tuple([self.pprint(v) for v in expression]))
                except KeyboardInterrupt as e:
                    raise e from e
                finally:
                    sys.stdout = s
        return str(expression)     
    
    def get_size(self):
        """
        retourne la largeur du terminal en nbr de caractere
        """
        return shutil.get_terminal_size((80, 20)).columns                       # avec un module python cette fois ci, pas besoin de protection car en cas d'echec, cette fonction renvoi 80

    def show(self, text, *, ecart=0, force=False):
        """
        affiche a l'ecran le message
        'ecart' est le nombre d'indentation a ajouter pour le prochain message, n'a pas d'effet dans l'affichage imediat
        'force' permet de forcer l'affichage meme si le nombre d'indentation maximum est atteind
        """
        assert type(text) is str, "'text' doit etre une chaine de caractere, pas un %s." % type(text)
        
        nbr_colones_avant = self.nbr_colones                                    # le nombre de colones avant l'affichage
        indentation = self.get_indentation(ecart=ecart)                         # le nombre d'indentation pour l'affichage la tout de suite
        nbr_colones_maintenant = self.nbr_colones                               # le nombre de colone au moment de l'affichage
        ma_colone = self.ma_colone                                              # le numero de la colone ou l'on doit affiher la tout de suite. Les numeros de colones partent de 0
        size = self.get_size()                                                  # largeur du terminal
        if indentation + min(0, ecart) <= self.display or force:                # si il faut afficher car la limite d'indentation n'est pas encore atteinte
            if nbr_colones_avant == 1:                                          # si il n'y a qu'une seule colone
                print("\n".join((("    "*indentation + l)[:size] for l in text.split("\n"))))# et bien on se gene pas
            else:                                                               # par contre, si il y a plusieurs colones
                for ligne in text.split("\n"):
                    debut = (" "*(size//nbr_colones_avant - 1)+"|")*ma_colone   # partie gauche du message
                    if ma_colone + 1 < nbr_colones_avant:
                        milieu = ("    "*indentation + ligne + " "*size)[:(size//nbr_colones_avant - 1)] # la partie centrale avec l'affichage
                        fin = ("|" + " "*(size//nbr_colones_avant-1))*(nbr_colones_avant-ma_colone-2) + "|"
                    else:
                        milieu = ""
                        fin = ("    "*indentation + ligne)[:(size//nbr_colones_avant - 1)]
                    print(debut + milieu + fin)
        
        if self.nbr_colones != nbr_colones_avant:
            return None
            raise NotImplementedError("La liaison et la suppression des colones n'est pas assuree")

    def __enter__(self):
        """
        permet de pouvoir etre appelle avec 'with'
        """
        self.show(self.text, ecart=1)
        return self

    def __exit__(self, type, value, traceback):
        """
        appelle l'orsque l'on sort du bloc 'with'
        """
        if (type, value, traceback) == (None, None, None):
            self.show("Success!", ecart=-1)
        else:
            erreur = str(value)
            if erreur:
                self.show("Failure! : %s" % erreur, ecart=-1)
            else:
                self.show("Failure!", ecart=-1)

class Str(str):
    """
    extension de l'objet 'str' primitatf a python
    """
    def __init__(self, *args, **kwargs):
        super().__init__()
        self = str(*args, **kwargs)

    def indice(self):
        """
        retourne une chaine ou tous les caracteres sont en indices
        en cas d'echec, retourne la chaine telle quelle etait
        """
        correspondance = {" ":" ", "\t":"\t",
            "0":"₀", "1":"₁", "2":"₂", "3":"₃", "4":"₄", "5":"₅", "6":"₆", "7":"₇", "8":"₈", "9":"₉",
            "a":"ₐ", "b":"ɂ", "c":"ɂ", "d":"ɂ", "e":"ₑ", "f":"ɂ", "g":"ɂ", "h":"ₕ", "i":"ᵢ", "j":"ⱼ", "k":"ₖ", "l":"ₗ", "m":"ₘ", "n":"ₙ", "o":"ₒ", "p":"ₚ", "q":"ɂ", "r":"ᵣ", "s":"ₛ", "t":"ₜ", "u":"ᵤ", "v":"ᵥ", "w":"ɂ", "x":"ₓ", "y":"ɂ", "z":"ɂ",
            "A":"ɂ", "B":"ʙ", "C":"ϲ", "D":"ɂ", "E":"ɂ", "F":"ɍ", "G":"ɢ", "H":"ʜ", "I":"ɪ", "J":"ȷ", "K":"κ", "L":"ʟ", "M":"м", "N":"ɴ", "O":"ο", "P":"ɂ", "Q":"ǫ", "R":"ʀ", "S":"ɂ", "T":"ɂ", "U":"ɂ", "V":"ɂ", "W":"ɯ", "X":"х", "Y":"ʏ", "Z":"ƶ",
            "(":"₍", ")":"₎", "+":"₊", "-":"₋", "=":"₌",
            "°":"˳", "~":"˷", ",":"ˏ"
            }
        
        chaine_indice = ""              # la chaine qui va etre renvoyee
        ind = [i for n,i in correspondance.items()]# liste qui contient tous les caractere en indices

        for car in self:                # pour chaqun des caractere de la chaine a metre en indice
            if car in correspondance:   # si on sait le convertir
                chaine_indice += correspondance[car]# on converti le caractere affin d'en donner ca representation en indice
            elif car in ind:            # si le caractere est deja un caractere en indice
                chaine_indice += car    # on le laisse tel qu'elle
            else:                       # si ce n'est ni l'un ni l'autre
                return self             # alors on ne sait pas faire!
        return Str(chaine_indice)       # si on sort de la boucle, c'est que on a bien reussi la convertion

    def exposant(self):
        """
        retourne la chaine en metant les caracteres en exposants
        en cas d'echec, retourne la chaine de caractere telle qu'elle etait sans rien changer
        """
        correspondance = {" ":" ", "\t":"\t",
            "0":"⁰", "1":"¹", "2":"²", "3":"³", "4":"⁴", "5":"⁵", "6":"⁶", "7":"⁷", "8":"⁸", "9":"⁹",
            "a":"?", "b":"?", "c":"?", "d":"?", "e":"?", "f":"?", "g":"?", "h":"?", "i":"?", "j":"?", "k":"?", "l":"?", "m":"?", "n":"?", "o":"?", "p":"?", "q":"?", "r":"?", "s":"?", "t":"?", "u":"?", "v":"?", "w":"?", "x":"?", "y":"?", "z":"?",
            "A":"?", "B":"?", "C":"?", "D":"?", "E":"?", "F":"?", "G":"?", "H":"?", "I":"?", "J":"?", "K":"?", "L":"?", "M":"?", "N":"?", "O":"?", "P":"?", "Q":"?", "R":"?", "S":"?", "T":"?", "U":"?", "V":"?", "W":"?", "X":"?", "Y":"?", "Z":"?",
            }
        
        chaine_exposant = ""            # la chaine qui va etre renvoyee
        ind = [i for n,i in correspondance.items()]# liste qui contient tous les caracteres en exposant

        for car in self:                # pour chaqun des caractere de la chaine a metre en exposant
            if car in correspondance:   # si on sait le convertir
                chaine_exposant += correspondance[car]# on converti le caractere affin d'en donner ca representation en exposant
            elif car in ind:            # si le caractere est deja un caractere en exposant
                chaine_exposant += car  # on le laisse tel qu'elle
            else:                       # si ce n'est ni l'un ni l'autre
                return self             # alors on ne sait pas faire!
        return Str(chaine_exposant)     # si on sort de la boucle, c'est que on a bien reussi la convertion

class Temprep:
    """
    cree puis detruit un dossier temporaire
    la representation de cet objet est le chemin absolu de ce repertoire
    """
    def __init__(self, destroy=True):
        self.destroy = destroy              # True imposse de supprimer le repertoir a la fin, False le laisse en vie
        self.temprep = None                 # creation du repertoire contenant tous les droits

    def __call__(self):
        if self.temprep == None:
            self.temprep = tempfile.mkdtemp()
        return self.temprep

    def __repr__(self):
        return self.__call__()

    def __str__(self):
        return self.__call__()

    def __del__(self):
        if self.destroy and self.temprep:
            try:
                shutil.rmtree(self.temprep)
            except KeyboardInterrupt as e:
                raise e from e
            except Exception as e:
                try:
                    import importlib
                    import shutil
                    importlib.reload(shutil)
                    shutil.rmtree(self.temprep)
                except Exception as e:
                    pass

class Translator:
    """
    permet de traduire du texte
    se base sur un dictionaire de la forme:
    {
    langue_source1:
        {
        message1:
            {
            langue_destination1: message_traduit1,
            langue_destination2: message_traduit2
            },
        message2:
            {
            langue_destination1: message_traduit1,
            langue_destination2: message_traduit2
            }
        },
    langue_source2:
        {
        message1:
            {
            langue_destination1: message_traduit1,
            langue_destination2: message_traduit2
            },
        message2:
            {
            langue_destination1: message_traduit1,
            langue_destination2: message_traduit2
            }
        }
    }
    """
    def __init__(self, src, dest):
        if os.path.exists(os.path.join(str(raisin.temprep), "translation_table.json")):# si on a un endroit ou stocker les resultats
            self.save = True                        # on retient qu'il faut les enregistrer pour la prochaine fois 
            with open(os.path.join(str(raisin.temprep), "translation_table.json"), "r") as f:# on recupere ceux qui sont deja stockee
                self.translation_table = json.load(f)# pour cela on utilise json qui est plus lisible
        else:                                       # si l'on a pas d'espace dedier
            self.save = False                       # on ne va pas se permetre d'en prendre un!
            self.translation_table = {}             # tous va rester dans la rame
        self.src = src                              # langue des messages a traduire
        if dest == "auto":                          # si on ne sait pas dans quelle langue traduire
            self.dest = self.get_local_language()   # on traduit dans la langue du pays ou l'on est
        else:                                       # si la langue de sortie est imposee
            self.dest = dest                        # on se pli aux voeux
        self.last = time.time()

    def __call__(self, message, src="default", dest="default"):
        """
        cette methode permet une traduction plus discrete
        """
        if src == "default":                        # si la source n'est pas imposee
            src = self.src                          # on prend la source par defautl
        if dest == "auto":                          # si la langue de destination doit etre adapte
            dest = self.get_local_language()        # on refait une requette
        elif dest == "default":                     # si par contre on demand de bien garder la meme langue de destination
            dest = self.dest                        # on ne fait pas le difficile
        return self.translate(message, src=src, dest=dest)

    def translate(self, message, src, dest):
        """
        traduit le message 'message'
        ce message est ecrit dans la langue 'src'
        ce message va etre traduit dans la langue 'dest'
        retourne le message traduit
        en cas d'echec, retourne le message d'origine
        """
        @timeout_decorator(5)
        def traduire_internet(message, src, dest):
            if googletrans:
                return googletrans.Translator().translate(message, src=src, dest=dest).text
            return message

        if src == "default":                        # si la source n'est pas imposee
            src = self.src                          # on prend la source par defautl
        if dest == "auto":                          # si la langue de destination doit etre adapte
            dest = self.get_local_language()        # on refait une requette
        elif dest == "default":                     # si par contre on demand de bien garder la meme langue de destination
            dest = self.dest                        # on ne fait pas le difficile

        if src == dest:                                                         # si le texte est deja dans la bonne langue
            return message                                                      # on ne fait pas vraiment d'effort
        if dest in self.translation_table.get(src, {}).get(message, {}):        # de meme, si il est deja traduit, on evite
            return self.translation_table[src][message][dest]                   # de le traduire a nouveau
        if not googletrans:                                                     # si googletrans n'est pas installe
            return message                                                      # on ne peut pas aller plus loin
        
        try:                                                                    # si il n'est pas traduit, on va le faire avec google traduction
            if src not in self.translation_table:                               # il faut donc d'abord preparer le terrain
                self.translation_table[src] = {}                                # en s'assurant que l'on puisse enregister le resultat au bon endroit
            if message not in self.translation_table[src]:                      # c'est pourquoi l'on verifie que
                self.translation_table[src][message] = {}                       # le chemin d'acces soit complet, on le cre si nesessaire
            self.translation_table[src][message][dest] = traduire_internet(message, src, dest)# c'est la qu'on peu enfin traduire le message
        except KeyboardInterrupt as e:                                          # si pandant la traduction, l'utilisateur en a marre
            raise e from e                                                      # on lui donne la prioritee
        except Exception as e:                                                  # si une autre erreur survient
            raise e from e
            return message                                                      # tant pis on abandonne la traduction

        self.save_ardware()                                                     # on enregistre ce changement affin de pouvoir en profiter la prochaine fois
        return self.translation_table[src][message][dest]                       # puis on retourne efin le message traduit

    def save_ardware(self):
        """
        ecrase le fichier json pour enregistrer un fichier plus recent
        """
        if not self.save:                           # si l'on a pas le droit d'enregistrer
            return None                             # on ne fait rien
        if time.time()-self.last < 10:              # si on a fait une mise a jour il y a moin de 10 secondes
            return None                             # on ne fait iren non plus
        self.last = time.time()                     # si il faut se metre au boulot, on le fait mais on s'aprette tout de meme a en faire le moin possible
        with open(os.path.join(str(raisin.temprep), "translation_table.json"), "w") as f:
            json.dump(f, self.translation_table)

    def get_local_language(self):
        """
        retourne les initiale de la langue locale
        """
        if "language" in self.translation_table:
            return self.translation_table["language"]
        if raisin.Id().country is not None:
            self.translation_table["language"] = raisin.Id().country
            return raisin.Id().country
        return "fr"

def timeout_decorator(seconds):
    """
    controle le temps d'excecution d'une fonction
    si le temps d'execution est superieur a 'seconds'
    retourne une exception
    """
    def decorateur(fonction_a_executer):
        """
        c'est lui qui est appelle directement lors de la definition de notre fonction
        c'est en fit lui le decorateur
        """
        def fonction_modifiee(*args, **kwargs):
            """
            c'est la fonction renvoyee par notre decorateur
            elle se charge de verifier le temps passe
            """
            class Timeouteur(threading.Thread):
                def __init__(self, function, *args, **kwargs):
                    threading.Thread.__init__(self)
                    self.function = function
                    self.args = args
                    self.kwargs = kwargs
                    self.valeur_renvoyee = None
                    self.resultat_ok = False        # True si un resultat a ete renvoye dans le temps imparti
                    self.erreur = None              # le message d'erreur si il y en a un
                
                def run(self):
                    try:
                        self.valeur_renvoyee = self.function(*self.args, **self.kwargs)
                        self.resultat_ok = True
                    except Exception as e:
                        self.erreur = e

            encapsulation = Timeouteur(fonction_a_executer, *args, **kwargs)
            encapsulation.start()
            encapsulation.join(timeout=seconds)
            if encapsulation.resultat_ok:
                valeur_renvoyee = encapsulation.valeur_renvoyee
                return valeur_renvoyee
            raise encapsulation.erreur
        
        return fonction_modifiee

    return decorateur

def get_temperature(dt=0):
    """
    |=================================|
    | Recupere la temperature du CPU. |
    |=================================|

    En cas d'echec, retourne None.

    parametres
    ----------
    :param dt: Duree en seconde de la mesure.
    :type dt: float

    sortie
    ------
    :return: La temperature moyenne en °C.
    :rtype: float

    exemple
    -------
    :Example:
    >>> from raisin.tools import get_temperature
    >>>
    >>> get_temperature()
    58.0
    >>>
    """
    def get():
        """
        Recupere la liste des temperature instentanement.
        """
        # Avec Psutil
        try:
            return [t.current for t in psutil.sensors_temperatures()["coretemp"]]
        except AttributeError:
            if os.name == "nt":
                p.show("Haha! t'as cru que microchiotte windaube sait lire la temperature! A ton avis, pourquoi tant de CPU grillent avec ce system?\n"\
                       "Plus serieusement, si tu tiens a cette option, passe sur un vrai systeme d'exploitation...")
                return None
            if psutil.__version__ < "5.1":
                p.show("Cette option est diponible a partir de la version 5.1, vous avez la %s. Vous pouvez taper: '%s -m pip install --upgrade psutil'." % (psutil.__version__, sys.executable))
            p.show("Cette option ne semble pas disponible, le module 'psutil' devrai avoir une fonction 'sensors_temperatures'. Or, ce n'est pas le cas.")
        
        # Sur raspberry
        try:
            with open("/opt/vc/bin/vcgencmd", "r") as ftemp:
                return [int(ftemp.read())/1000]
        except (FileNotFoundError, PermissionError):
            pass
        
        # Sur linux
        temp = []
        for file in glob.iglob("/sys/class/thermal/thermal_zone*/temp"):
            try:
                with open(file, "r") as ftemp:
                    temp.append(int(ftemp.read())/1000)
            except (FileNotFoundError, PermissionError):
                continue
        return temp if temp else None
    
    assert type(dt) in (float, int), \
        "'dt' doit etre un entier ou un flotant, pas un %s." % type(dt)
    with Printer("Get the CPU temperature...") as p:
        try:
            import psutil
        except ImportError:
            p.show("'psutil' est nessecaire pour pouvoir recuperer la temperature.")
            return None
        t_init = time.time()
        temperatures = get()
        if not temperatures:
            return None
        else:
            while t_init + dt > time.time():
                time.sleep(0.08)
                temperatures.extend(get())
            
            p.show("°C, ".join(map(str, temperatures[-16:])) + "°C")
            longueur = len(temperatures)
            if longueur % 2:
                temp = sorted(temperatures)[longueur//2]
            temperatures.sort()
            temp = (temperatures[longueur//2-1] + temperatures[longueur//2])/2
            return temp

def improve_recursion_decorateur(increment):
    """
    permet d'eviter les problemes de limites de recursion
    car si il y a un probleme, la limite est augmentee de 'increment'
    """
    assert type(increment) is int, "'increment' doit etre un entier, '%s' n'est pas un entier, c'est un %s." % (increment, type(increment))
    assert increment > 0, "'increment' doit etre strictement positif, ce n'est pas le cas de %d." % increment

    def decorateur(fonction_a_executer):

        def fonction_modifiee(*args, **kwargs):
            args_copie = copy.copy(args)
            kwargs_copie = copy.copy(kwargs)
            while 1:
                try:
                    return fonction_a_executer(*args_copie, **kwargs_copie)
                except RecursionError as e:
                    if sys.getrecursionlimit() >= 10000:                                    # si on a deja depasse une limite raisonable de recursion
                        raise e from e                                                      # on est surement dans une boucle infinie, donc on ne s'entete pas
                    try:
                        avant = sys.getrecursionlimit()
                        sys.setrecursionlimit(min(10000, avant + increment))                # si il y a trop de recursions par exemple a + b + c - g * 3 + ...
                    except Exception as e:
                        raise e from e
                    logging.warning("Recursion limit increased from %d to %d." % (avant, sys.getrecursionlimit()))
                    continue
                except Exception as e:
                    raise e from e
        return fonction_modifiee
    return decorateur

id_ = Id()
temprep = Temprep()



#!/usr/bin/env python3

"""
|=================================|
| Permet d'executer des fonctions |
| en ligne de commandes.          |
|=================================|

Permet en particulier:
1) Installer/Desinstaller raisin ou un autre module.
    -Installer raisin:
        $ python3 -m raisin install
    -Desinstaller raisin:
        $ python3 -m raisin uninstall
2) Configurer raisin.
    -Configuration generale:
        $ python3 -m raisin configure
    -Changer le mot de passe:
        $ python3 -m raisin psw
        $ python3 -m raisin configure psw
    -Configurer l'antivol:
        $ python3 -m raisin padlock
        $ python3 -m raisin configure padlock
"""

import argparse
import getpass
import os
import sys

import raisin


def parse_arguments():
    """
    Fait le parsing des arguments.
    """
    parser = argparse.ArgumentParser(description="Simple API pour raisin.")
    subparsers = parser.add_subparsers(dest="parser_name")

    # install
    parser_install = subparsers.add_parser("install", help="Installer un module python, possiblement raisin lui-meme.")
    parser_install.add_argument("module", type=str, nargs="?", default="raisin", help="Nom du module")
    parser_install.add_argument("-U", "--upgrade", action="store_true", default=False, help="Fait la mise a jour.")

    # uninstall
    parser_uninstall = subparsers.add_parser("uninstall", help="Desinstaller un module python, possiblement raisin lui-meme.")
    parser_uninstall.add_argument("module", type=str, nargs="?", default="raisin", help="Nom du module.")

    # configuration
    parser_configure = subparsers.add_parser("configure", help="Personnaliser l'installation de raisin.")
    parser_configure.add_argument("category", type=str, nargs="?", choices=["all", "psw", "padlock"], default="all", help="Specifier l'element a configurer.")

    # psw
    parser_psw = subparsers.add_parser("psw", help="Ajouter/Changer/Supprimer le mot de passe de raisin.")

    # padlock
    parser_padlock = subparsers.add_parser("padlock", help="Ajouter/Changer/Desactiver l'antivol de raisin.")

    parser_start = subparsers.add_parser("start", help="Lancer les utilitaires de raisin.")
    parser_start.add_argument("name", type=str, nargs="?", default="all", choices=["all", "server", "padlock", "statistics"], help="Nom de l'application a lancer.")

    parser_cipher = subparsers.add_parser("cipher", help="Chiffrer un fichier ou un dossier.")
    parser_cipher.add_argument("filename", type=str, nargs="+", help="Nom du fichier a chiffrer.")
    parser_cipher.add_argument("-p", "-psw", "--psw", "--password", type=str, nargs="?", default="", help="Permet d'entrer un mot de passe, s'il il est omis, la clef publique est utilisee.")

    parser_uncipher = subparsers.add_parser("uncipher", help="Dechiffrer un fichier ou un dossier.")
    parser_uncipher.add_argument("filename", type=str, nargs="+", help="Nom du fichier a chiffrer.")
    parser_uncipher.add_argument("-p", "-psw", "--psw", "--password", type=str, nargs="?", default="", help="Permet d'entrer un mot de passe, sil il est omis, la clef privee est utilisee.")

    return parser

def _install_raisin():
    import raisin.application.install as install
    return install.main()

def _uninstall_raisin():
    import raisin.application.uninstall as uninstall
    return uninstall.main()

def _configure_raisin():
    import raisin.application.configuration as configuration
    configuration.Config()
    return 0

def _psw_raisin():
    import raisin.security as security
    security.change_psw()
    return 0

def _padlock_raisin():
    import raisin.application.configuration as configuration
    configuration.change_padlock()
    return 0

def main(args_brut=[]):
    """
    Retourne 0 en cas de success.
    'args_brut' permet de pouvoir passer des commandes plus facilement
    depuis un script directement.
    """
    parser = parse_arguments()
    if args_brut:
        args = parser.parse_args(args_brut)
    else:
        args = parser.parse_args()
    
    if args.parser_name == "install":
        if args.module == "raisin":
            if args.upgrade:
                raise NotImplementedError("Impossible pour le moment de metre raisin a jour.")
            else:
                return _install_raisin()
        else:
            raise NotImplementedError("Impossible d'installer un aute module que raisin.")
    elif args.parser_name == "uninstall":
        if args.module == "raisin":
            return _uninstall_raisin()
        else:
            raise NotImplementedError("Je ne suis pas capable de desinstaller un module autre que raisin.")
    elif args.parser_name == "configure":       # si il faut configurer raisin
        if args.category == "all":
            return _configure_raisin()
        elif args.category == "psw":
            return _psw_raisin()
        elif args.category == "padlock":
            return _padlock_raisin()
    elif args.parser_name == "psw":
        return _psw_raisin()
    elif args.parser_name == "padlock":
        return _padlock_raisin()

    elif args.parser_name == "start":           # si il faut lancer des scripts specifiques
        tout = True if args.name == "all" else False # booleen qui dit si il faut tout lancer ou pas
        raise NotImplementedError("Impossible de lancer l'application.")
    elif args.parser_name in ("cipher", "uncipher"):
        raise NotImplementedError("Impossible de chiffrer/dechifrer un fichier.")
    else:
        sys.stderr.write("Aucun arguments n'est passe en parametre.\nPour plus d'information, tapez '%s -m raisin --help'\n" % sys.executable)
        return 1

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
|==========================================|
| Demande et modification du mot de passe. |
|==========================================|

Interagit avec l'utilisateur afin de lui demander le mot de passe.
C'est aussi ici qu'il y a l'interface qui permet de changer
le mot de passe.
"""

from ...tools import Printer
from ...security import request_psw
from ..hmi import tkinter, messagebox
from ..settings import settings
from .theme import *
import raisin.application.hmi.dialog as dialog


def get_new_psw(*, existing_window=None):
    """
    |===================================================|
    | Interface pour changer le mot de passe de raisin. |
    |===================================================|

    Utilise dans la mesure du possible, une jolie interface graphique.
    Les verifications sont faites ici.

    parametres
    ----------
    :param existing_window: Une fenetre tkinter presente par ailleur.

    sortie
    ------
    :return:
        -Le nouveau mot de passe.
        -La nouvelle sentence memory.
        -Si il faut enregistrer le mot de passe en clair.
    :rtype: (str, str, bool)
    """
    class DialogWindow:
        """
        Permet d'interragir graphiquement avec l'utilisateur.
        """
        def __init__(self, old_psw, sentence_memory, clair, existing_window):
            self.sentence_memory = sentence_memory
            self.clair = clair
            self.existing_window = existing_window
            self.violently_closed = True        # permet de savoir si la fenetre a ete fermee violement ou via le bouton 'valider'
            self.answer = None                  # reponse de l'utilisateur

            # preparation de la fenetre
            if self.existing_window:                                                        # si il y a deja une fenetre ouverte
                self.window = tkinter.Toplevel(self.existing_window)                        # on est oblige d'utiliser un toplevel sinon ca plante
                self.window.grab_set()                                                      # on fige la fenetre parente
                self.window.protocol("WM_DELETE_WINDOW", lambda : (self.window.destroy(), self.window.quit()))# il se trouve que ca semble fonctionner comme ca...
            else:                                                                           # dans le cas ou aucune instance tkinter n'existe
                self.window = tkinter.Tk()                                                  # et bien on en cre une tout simplement

            # configuration
            self.window.title("Change password")
            self.window.configure(background=JAUNE)
            self.window.columnconfigure(0, weight=1)                                         # numero de colone, etirement relatif: On rend l'onglet redimenssionable sur la largeur
            self.window.columnconfigure(1, weight=31)
            self.window.columnconfigure(2, weight=1)
            for i in range(6):
                self.window.rowconfigure(i, weight=1)
            self.window.focus_force()
            self.window.bind("<Return>", lambda event: self.quit())
            self.window.bind("<Escape>", lambda event: self.quit())

            # initialisation des variables
            self.psw1 = tkinter.StringVar()
            self.psw1.set(old_psw if old_psw is not None else "")
            self.psw2 = tkinter.StringVar()
            self.psw2.set(old_psw if old_psw is not None else "")
            self.ok_var = tkinter.StringVar()
            self.en_clair = tkinter.IntVar()
            self.en_clair.set(clair)
            self.memory = tkinter.StringVar()
            self.memory.set(sentence_memory)

            # remplissage + ecoute
            self.create_widgets()
            self.compare()
            self.window.mainloop()

        def create_widgets(self):
            """
            Initialisation des widgets.
            """
            theme(tkinter.Label(self.window, text="New Password :")).grid(row=0, column=0, sticky="w")
            entre1 = theme(tkinter.Entry(self.window, show="*", textvariable=self.psw1))
            entre1.bind("<KeyRelease>", lambda event : self.compare())
            entre1.grid(row=0, column=1, sticky="ew")
            theme(tkinter.Label(self.window, text="Confirmation :")).grid(row=1, column=0, sticky="w")
            entre2 = theme(tkinter.Entry(self.window, show="*", textvariable=self.psw2))
            entre2.bind("<KeyRelease>", lambda event : self.compare())
            entre2.grid(row=1, column=1, sticky="ew")
            self.canvas_ok = theme(tkinter.Canvas(self.window))
            self.canvas_ok.grid(row=2, column=0)
            theme(tkinter.Label(self.window, textvariable=self.ok_var)).grid(row=2, column=1)
            theme(tkinter.Checkbutton(self.window, variable=self.en_clair, text="Enregistrer le mot de passe en clair")).grid(row=3, column=0, columnspan=2, sticky="ew")
            theme(tkinter.Button(self.window,
                image=icons.info(),
                command=lambda: messagebox.showinfo(
                    "Info psw",
                    "Enregistrer le mot de passe en clair permet de donner de l'autonomie a l'application. "
                    "En effet, raisin sera capable tous seul d'aller chercher le mot de passe "
                    "sans avoir a vous le demander a tous bout de champs. Mais comme vous pouvez "
                    "vous en douter, cette option n'est pas ultra sécure!")
                )).grid(row=3, column=3)
            theme(tkinter.Label(self.window, text="sentence_memory :")).grid(row=4, column=0, sticky="w")
            theme(tkinter.Entry(self.window, textvariable=self.memory)).grid(row=4, column=1, sticky="ew")
            self.valider = theme(tkinter.Button(self.window, text="Valider", command=lambda: self.quit()))
            self.valider.grid(row=5, column=0, columnspan=2)

        def compare(self):
            """
            Compare les 2 mots de passes et fait apparaitre
            le bouton de validation.
            """
            self.canvas_ok.delete("all")
            if self.psw1.get() == self.psw2.get():
                self.canvas_ok.create_image(8, 8, image=icons.ok())
                self.ok_var.set("Passwords are the same.")
                self.valider.config(state="normal")
            else:
                self.canvas_ok.create_image(8, 8, image=icons.error())
                self.ok_var.set("Passwords do not match.")
                self.valider.config(state="disable")

        def quit(self):
            if self.psw1.get() != self.psw2.get():
                Printer().show("Passwords do not match.")
            else:
                self.violently_closed = False                                                   # si on ferme la fenetre depuis cette methode, c'est que la fermeture est propre
                self.answer = (self.psw1.get(), self.memory.get(), self.en_clair.get())                                                            # on peut allos recuperer la reponse
                self.window.destroy()                                                           # selon qu'il y ai deja une fenetre en arriere plan
                if self.existing_window:                                                        # on applique ou non la methode destroy
                    self.window.quit() 

    class DialogTerminal:
        """
        Permet d'interagir avec l'utilisateur dans le terminal.
        """
        def __init__(self, psw, sentence_memory, clair):
            self.psw = psw
            self.sentence_memory = sentence_memory
            self.clair = clair
            self.menu()

        def menu(self):
            """
            Choisi entre les differentes options.
            """
            choix1 = "Change / Delete the password." if self.psw else "Add a password."
            choix2 = "Change / Delete the mnemonic frase." if self.sentence_memory else "Add mnemonic frase."
            choix3 = "Do not save the password in clear text." if self.clair else "Save the password in clear text."
            rep = dialog.question_choix_exclusif(
                "What do you want to do?",
                choix=[
                    choix1,
                    choix2,
                    choix3,
                    "Apply changes and exit."
                ],
                default=3)
            if rep == 0:
                return self.get_psw()
            elif rep == 1:
                return self.get_mnemonic()
            elif rep == 2:
                self.clair = not self.clair
                return self.menu()

        def get_psw(self):
            """
            Supprime, change ou ajoute un mot de passe.
            """
            def get():
                psw = dialog.question_reponse("New password:", show="")
                dialog.question_reponse("Confirmation:", show="", validatecommand=lambda answer: answer == psw)
                return psw

            if not self.psw:
                psw = get()
                if psw:
                    self.psw = psw
                elif psw == "":
                    self.psw = None
            else:
                self.psw = None
                if dialog.question_choix_exclusif("Do you want to delete or change the password?", ["Delete.", "Change"]):
                    return self.get_psw()
            return self.menu()

        def get_mnemonic(self):
            """
            Demande une phrase mnemonique.
            """
            self.sentence_memory = dialog.question_reponse("mnemonic frase:")
            return self.menu()

    with Printer("Configure the raisin psw..."):
        old_psw = request_psw(force=True, message="Change the raisin psw.")
        sentence_memory = settings["account"]["security"]["sentence_memory"]
        clair = True if settings["account"]["security"]["psw"] else False
        
        if tkinter:
            g = DialogWindow(old_psw, sentence_memory, clair, existing_window=existing_window)
            if g.violently_closed:
                raise KeyboardInterrupt("La fenetre a ete fermee violement!")
            return g.answer
        else:
            g = DialogTerminal(old_psw, sentence_memory, clair)
            return g.psw, g.sentence_memory, g.clair
